@@ -1,7 +1,7 @@
 #!/usr/bin/env perl
 
 # docx2txt, a command-line utility to convert Docx documents to text format.
-# Copyright (C) 2008-2012 Sandeep Kumar
+# Copyright (C) 2008-2014 Sandeep Kumar
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -85,6 +85,20 @@
 #                 Usage is extended to accept docx file from standard input.
 #                 "-h" has to be given as the first argument to get usage help.
 #                 Added new configuration variable "config_tempDir".
+#    14/03/2014 - Remove deleted text from output. This effects in case changes
+#                 are being tracked in docx document. Patch was contributed by
+#                 William Parsons (wbparsons>AT<cshore>DOT<com).
+#                 Removed experimental config option config_exp_extra_deEscape.
+#    27/03/2014 - Remove non-document_text content marked by wp/wp14 tags.
+#    07/04/2014 - Added support for handling lists (bullet, decimal, letter,
+#                 roman) along with (attempt at) indentation.
+#                 Added new configuration variable config_twipsPerChar.
+#                 Removed configuration variable config_listIndent.
+#    14/04/2014 - Fixed list numbering - lvl start value needs to be considered.
+#                 Improved list indentation and corresponding code.
+#    27/04/2014 - Improved paragraph content layout/indentation.
+#    13/05/2014 - Added new configuration variable config_unzip_opts. Users can
+#                 now use unzipping programs like 7z, pkzipc, winzip as well.
 #
 
 
@@ -94,19 +108,13 @@
 #
 
 our $config_unzip = '/usr/bin/unzip';	# Windows path like 'C:/path/to/unzip.exe'
+our $config_unzip_opts = '-p';		# To extract file on standard output
 
 our $config_newLine = "\n";		# Alternative is "\r\n".
-our $config_listIndent = "  ";		# Indent nested lists by "\t", " " etc.
 our $config_lineWidth = 80;		# Line width, used for short line justification.
 our $config_showHyperLink = "N";	# Show hyperlink alongside linked text.
 our $config_tempDir;			# Directory for temporary file creation.
-
-
-#
-# Some experimental settings.
-#
-
-our $config_exp_extra_deEscape = "N";   # Extra conversion of &...; sequences.
+our $config_twipsPerChar = 120;		# Approx mapping for layout purpose.
 
 
 #
@@ -133,12 +141,6 @@ if ($ENV{OS} =~ /^Windows/ && !(exists $ENV{OSTYPE} || exists $ENV{HOME})) {
 }
 
 
-#
-# ToDo: Better list handling. Currently assumed 8 level nesting.
-#
-my @levchar = ('*', '+', 'o', '-', '**', '++', 'oo', '--');
-
-
 #
 # Character conversion tables
 #
@@ -240,6 +242,10 @@ my %splchars = (
 	"\xA0" => '!=',		# <neq>
 	"\xA4" => '<=',		# <leq>
 	"\xA5" => '>=',		# <geq>
+    },
+
+    "\xEF\x82" => {
+	"\xB7" => '*'		# small white square
     }
 );
 
@@ -348,6 +354,24 @@ sub readFileInto {
     close $fh;
 }
 
+sub readOptionalFileInto {
+    local $/ = undef;
+
+    stat("$_[0]");
+    if (-f _) {
+        if (-r _ && -T _) {
+            open my $fh, "$_[0]" or die "Couldn't read file <$_[0]>!\n";
+            binmode $fh;
+            $_[1] = <$fh>;
+            close $fh;
+        }
+        else {
+            die "Invalid <$_[0]>!\n";
+        }
+    }
+}
+
+
 
 #
 # Check whether first argument is specifying a directory holding extracted
@@ -377,10 +401,12 @@ else {
 # Extract xml document content from argument docx file/directory.
 #
 
+my $unzip_cmd = "'$config_unzip' $config_unzip_opts";
+
 if ($inpIsDir eq 'y') {
     readFileInto("$ARGV[0]/word/document.xml", $content);
 } else {
-    $content = `"$config_unzip" -p "$ARGV[0]" word/document.xml 2>$nullDevice`;
+    $content = `$unzip_cmd "$ARGV[0]" word/document.xml 2>$nullDevice`;
 }
 
 cleandie "Failed to extract required information from <$inputFileName>!\n" if ! $content;
@@ -412,7 +438,7 @@ binmode $txtfile;    # Ensure no auto-conversion of '\n' to '\r\n' on Windows.
 if ($inpIsDir eq 'y') {
     readFileInto("$ARGV[0]/word/_rels/document.xml.rels", $_);
 } else {
-    $_ = `"$config_unzip" -p "$ARGV[0]" word/_rels/document.xml.rels 2>$nullDevice`;
+    $_ = `$unzip_cmd "$ARGV[0]" word/_rels/document.xml.rels 2>$nullDevice`;
 }
 
 my %docurels;
@@ -421,6 +447,54 @@ while (/<Relationship Id="(.*?)" Type=".*?\/([^\/]*?)" Target="(.*?)"( .*?)?\/>/
     $docurels{"$2:$1"} = $3;
 }
 
+#
+# Gather list numbering information.
+#
+
+$_ = "";
+if ($inpIsDir eq 'y') {
+    readOptionalFileInto("$ARGV[0]/word/numbering.xml", $_);
+} else {
+    $_ = `$unzip_cmd "$ARGV[0]" word/numbering.xml 2>$nullDevice`;
+}
+
+my %abstractNum;
+my @N2ANId = ();
+
+my %NFList = (
+    "bullet"      => \&bullet,
+    "decimal"     => \&decimal,
+    "lowerLetter" => \&lowerLetter,
+    "upperLetter" => \&upperLetter,
+    "lowerRoman"  => \&lowerRoman,
+    "upperRoman"  => \&upperRoman
+);
+
+if ($_) {
+    while (/<w:abstractNum w:abstractNumId="(\d+)">(.*?)<\/w:abstractNum>/g)
+    {
+        my $abstractNumId = $1, $temp = $2;
+
+	while ($temp =~ /<w:lvl w:ilvl="(\d+)"[^>]*><w:start w:val="(\d+)"[^>]*><w:numFmt w:val="(.*?)"[^>]*>.*?<w:lvlText w:val="(.*?)"[^>]*>.*?<w:ind w:left="(\d+)" w:hanging="(\d+)"[^>]*>/g )
+        {
+            # $2: Start $3: NumFmt, $4: LvlText, ($5,$6): (Indent (twips), hanging)
+
+            @{$abstractNum{"$abstractNumId:$1"}} = (
+                $NFList{$3},
+                $4,
+                $2,
+                int ((($5-$6) / $config_twipsPerChar) + 0.5),
+                $5
+            );
+        }
+    }
+
+    while ( /<w:num w:numId="(\d+)"><w:abstractNumId w:val="(\d+)"/g )
+    {
+        $N2ANId[$1] = $2;
+    }
+}
+
 # Remove the temporary file (if) created to store input from STDIN. All the
 # (needed) data is read from it already.
 unlink("$tempFile") if -e "$tempFile";
@@ -457,6 +531,125 @@ sub hyperlink {
     return $hltext;
 }
 
+#
+# Subroutines for processing numbering information.
+#
+
+my @RomanNumbers = ( "",
+    "i", "ii", "iii", "iv", "v", "vi", "vii", "viii", "ix", "x", "xi", "xii",
+    "xiii", "xiv", "xv", "xvi", "xvii", "xviii", "xix", "xx", "xxi", "xxii",
+    "xxiii", "xxiv", "xxv", "xxvi", "xxvii", "xxviii", "xxix", "xxx", "xxxi",
+    "xxxii", "xxxiii", "xxxiv", "xxxv", "xxxvi", "xxxvii", "xxxviii", "xxxix",
+    "xl", "xli", "xlii", "xliii", "xliv", "xlv", "xlvi", "xlvii", "xlviii",
+    "xlix", "l", "li" );
+
+
+sub lowerRoman {
+    return $RomanNumbers[$_[0]] if ($_[0] < @RomanNumbers);
+
+    @rcode = ("i", "iv", "v", "ix", "x", "xl", "l", "xc", "c", "cd", "d", "cm", "m");
+    @dval = (1, 4, 5, 9, 10, 40, 50, 90, 100, 400, 500, 900, 1000);
+
+    my $roman = "";
+    my $num = $_[0];
+
+    my $div, $i = (@rcode - 1);
+    while ($num > 0) {
+        $i-- while ($num < $dval[$i]);
+        $div = $num / $dval[$i];
+        $num = $num % $dval[$i];
+        $roman .= $rcode[$i] x $div;
+    }
+
+    return $roman;
+}
+
+sub upperRoman {
+    return uc lowerRoman(@_);
+}
+
+
+sub lowerLetter {
+    @Alphabets = split '' , "abcdefghijklmnopqrstuvwxyz";
+    return $Alphabets[($_[0] % 26) - 1] x (($_[0] - 1)/26 + 1);
+}
+
+sub upperLetter {
+    return uc lowerLetter(@_);
+}
+
+
+sub decimal {
+    return $_[0];
+}
+
+
+my %bullets = (
+    "\x6F" => 'o',
+    "\xEF\x81\xB6" => '::',	# Diamond
+    "\xEF\x82\xA7" => '#',	# Small Black Square
+    "\xEF\x82\xB7" => '*',	# Small Black Circle
+    "\xEF\x83\x98" => '>',	# Arrowhead
+    "\xEF\x83\xBC" => '+'	# Right Sign
+);
+
+sub bullet {
+    return $bullets{$_[0]} ? $bullets{$_[0]} : 'oo';
+}
+    
+my @lastCnt = (0);
+my @twipStack = (0);
+my @keyStack = (undef);
+my $ssiz = 1;
+
+sub listNumbering {
+    my $aref = \@{$abstractNum{"$N2ANId[$_[0]]:$_[1]"}};
+    my $lvlText;
+
+    if ($aref->[0] != \&bullet) {
+        my $key = "$N2ANId[$_[0]]:$_[1]";
+        my $ccnt;
+
+        if ($aref->[4] < $twipStack[$ssiz-1]) {
+            while ($twipStack[$ssiz-1] > $aref->[4]) {
+                pop @twipStack;
+                pop @keyStack;
+                pop @lastCnt;
+                $ssiz--;
+            }
+        }
+
+        if ($aref->[4] == $twipStack[$ssiz-1]) {
+            if ($key eq $keyStack[$ssiz-1]) {
+                ++$lastCnt[$ssiz-1];
+            }
+            else {
+                $keyStack[$ssiz-1] = $key;
+                $lastCnt[$ssiz-1] = $aref->[2];
+            }
+        }
+        elsif ($aref->[4] > $twipStack[$ssiz-1]) {
+            push @twipStack, $aref->[4];
+            push @keyStack, $key;
+            push @lastCnt, $aref->[2];
+            $ssiz++;
+        }
+
+        $ccnt = $lastCnt[$ssiz-1];
+
+        $lvlText = $aref->[1];
+        $lvlText =~ s/%\d([^%]*)$/($aref->[0]->($ccnt)).$1/oe;
+
+        my $i = $ssiz - 2;
+        $i-- while ($lvlText =~ s/%\d([^%]*)$/$lastCnt[$i]$1/o);
+    }
+    else {
+        $lvlText = $aref->[0]->($aref->[1]);
+    }
+
+    return ' ' x $aref->[3] . $lvlText . ' ';
+}
+
 #
 # Subroutines for processing paragraph content.
 #
@@ -479,9 +672,11 @@ my %tag2chr = (tab => "\t", noBreakHyphen => "-", softHyphen => " - ");
 
 $content =~ s/<?xml .*?\?>(\r)?\n//;
 
-# Remove the field instructions (instrText) and data (fldData).
-$content =~ s|<w:instrText[^>]*>.*?</w:instrText>||og;
-$content =~ s|<w:fldData[^>]*>[^<]*?</w:fldData>||og;
+$content =~ s{<(wp14|wp):[^>]*>.*?</\1:[^>]*>}||og;
+
+# Remove the field instructions (instrText) and data (fldData), and deleted
+# text.
+$content =~ s{<w:(instrText|fldData|delText)[^>]*>.*?</w:\1>}||ogs;
 
 # Mark cross-reference superscripting within [...].
 $content =~ s|<w:vertAlign w:val="superscript"/></w:rPr><w:t>(.*?)</w:t>|[$1]|og;
@@ -491,23 +686,18 @@ $content =~ s{<w:(tab|noBreakHyphen|softHyphen)/>}|$tag2chr{$1}|og;
 my $hr = '-' x $config_lineWidth . $config_newLine;
 $content =~ s|<w:pBdr>.*?</w:pBdr>|$hr|og;
 
-$content =~ s|<w:numPr><w:ilvl w:val="([0-9]+)"/>|$config_listIndent x $1 . "$levchar[$1] "|oge;
-
-#
-# Uncomment either of below two lines and comment above line, if dealing
-# with more than 8 level nested lists.
-#
-
-# $content =~ s|<w:numPr><w:ilvl w:val="([0-9]+)"/>|$config_listIndent x $1 . '* '|oge;
-# $content =~ s|<w:numPr><w:ilvl w:val="([0-9]+)"/>|'*' x ($1+1) . ' '|oge;
-
 $content =~ s{<w:caps/>.*?(<w:t>|<w:t [^>]+>)(.*?)</w:t>}/uc $2/oge;
 
 $content =~ s{<w:hyperlink r:id="(.*?)".*?>(.*?)</w:hyperlink>}/hyperlink($1,$2)/oge;
 
-$content =~ s/<w:p[^>]+?>(.*?)<\/w:p>/processParagraph($1)/oge;
+$content =~ s|<w:numPr><w:ilvl w:val="(\d+)"/><w:numId w:val="(\d+)"\/>|listNumbering($2,$1)|oge;
+
+$content =~ s{<w:ind w:(left|firstLine)="(\d+)"( w:hanging="(\d+)")?[^>]*>}|' ' x int((($2-$4)/$config_twipsPerChar)+0.5)|oge;
+
+$content =~ s{<w:p [^/>]+?/>|<w:br/>}|$config_newLine|og;
+
+$content =~ s/<w:p[^>]+?>(.*?)<\/w:p>/processParagraph($1)/ogse;
 
-$content =~ s{<w:p [^/>]+?/>|</w:p>|<w:br/>}|$config_newLine|og;
 $content =~ s/<.*?>//og;
 
 
@@ -515,19 +705,13 @@ $content =~ s/<.*?>//og;
 # Convert non-ASCII characters/character sequences to ASCII characters.
 #
 
-$content =~ s/(\xC2|\xC3|\xCF|\xE2.)(.)/($splchars{$1}{$2} ? $splchars{$1}{$2} : $1.$2)/oge;
+$content =~ s/(\xC2|\xC3|\xCF|\xE2.|\xEF.)(.)/($splchars{$1}{$2} ? $splchars{$1}{$2} : $1.$2)/oge;
 
 #
 # Convert docx specific (reserved HTML/XHTML) escape characters.
 #
 $content =~ s/(&)(amp|apos|gt|lt|quot)(;)/$escChrs{lc $2}/iog;
 
-#
-# Another pass for experimental text experience, after sequences like
-# "&amp;laquo;" are converted to "&laquo;".
-#
-$content =~ s/((&)([a-z]+)(;))/($escChrs{lc $3} ? $escChrs{lc $3} : $1)/ioge if (lc $config_exp_extra_deEscape eq "y");
-
 #
 # Write the extracted and converted text contents to output.
 #
